package sun.misc;

import java.util.NoSuchElementException;
import java.util.Iterator;
import java.util.AbstractSet;
import java.lang.ref.SoftReference;
import java.util.HashMap;
import java.util.Set;
import java.lang.ref.ReferenceQueue;
import java.util.Map;
import java.util.AbstractMap;

public class SoftCache extends AbstractMap implements Map
{
    private Map hash;
    private ReferenceQueue queue;
    private Set entrySet;
    
    private void processQueue() {
        SoftCache.ValueCell valueCell;
        while ((valueCell = (SoftCache.ValueCell)this.queue.poll()) != null) {
            if (SoftCache.ValueCell.access$000(valueCell)) {
                this.hash.remove(valueCell.key);
            }
            else {
                SoftCache.ValueCell.dropped--;
            }
        }
    }
    
    public SoftCache(int initialCapacity, float loadFactor) {
        this.queue = new ReferenceQueue();
        this.entrySet = null;
        this.hash = new HashMap(initialCapacity, loadFactor);
    }
    
    public SoftCache(int initialCapacity) {
        this.queue = new ReferenceQueue();
        this.entrySet = null;
        this.hash = new HashMap(initialCapacity);
    }
    
    public SoftCache() {
        this.queue = new ReferenceQueue();
        this.entrySet = null;
        this.hash = new HashMap();
    }
    
    @Override
    public int size() {
        return this.entrySet().size();
    }
    
    @Override
    public boolean isEmpty() {
        return this.entrySet().isEmpty();
    }
    
    @Override
    public boolean containsKey(Object o) {
        return SoftCache.ValueCell.access$300(this.hash.get(o), false) != null;
    }
    
    protected Object fill(Object o) {
        return null;
    }
    
    @Override
    public Object get(Object o) {
        this.processQueue();
        Object o2 = this.hash.get(o);
        if (o2 == null) {
            o2 = this.fill(o);
            if (o2 != null) {
                this.hash.put(o, SoftCache.ValueCell.access$400(o, o2, this.queue));
                return o2;
            }
        }
        return SoftCache.ValueCell.access$300(o2, false);
    }
    
    @Override
    public Object put(Object o, Object o2) {
        this.processQueue();
        return SoftCache.ValueCell.access$300(this.hash.put(o, SoftCache.ValueCell.access$400(o, o2, this.queue)), true);
    }
    
    @Override
    public Object remove(Object o) {
        this.processQueue();
        return SoftCache.ValueCell.access$300(this.hash.remove(o), true);
    }
    
    @Override
    public void clear() {
        this.processQueue();
        this.hash.clear();
    }
    
    private static boolean valEquals(Object o, Object obj) {
        return (o == null) ? (obj == null) : o.equals(obj);
    }
    
    @Override
    public Set entrySet() {
        if (this.entrySet == null) {
            this.entrySet = new SoftCache.EntrySet();
        }
        return this.entrySet;
    }
    
    static boolean access$600(Object o, Object o2) {
        return valEquals(o, o2);
    }
    
    static void access$800(SoftCache softCache) {
        softCache.processQueue();
    }
    
    private static class ValueCell extends SoftReference
    {
        private static Object INVALID_KEY;
        private static int dropped;
        private Object key;
        
        private ValueCell(Object key, Object referent, ReferenceQueue q) {
            super(referent, q);
            this.key = key;
        }
        
        private static SoftCache.ValueCell create(Object o, Object o2, ReferenceQueue referenceQueue) {
            if (o2 == null) {
                return null;
            }
            return new SoftCache.ValueCell(o, o2, referenceQueue);
        }
        
        private static Object strip(Object o, boolean b) {
            if (o == null) {
                return null;
            }
            SoftCache.ValueCell valueCell = (SoftCache.ValueCell)o;
            Object value = valueCell.get();
            if (b) {
                valueCell.drop();
                return value;
            }
            return value;
        }
        
        private boolean isValid() {
            return this.key != SoftCache.ValueCell.INVALID_KEY;
        }
        
        private void drop() {
            super.clear();
            this.key = SoftCache.ValueCell.INVALID_KEY;
            ++SoftCache.ValueCell.dropped;
        }
        
        static boolean access$000(SoftCache.ValueCell valueCell) {
            return valueCell.isValid();
        }
        
        static Object access$300(Object o, boolean b) {
            return strip(o, b);
        }
        
        static SoftCache.ValueCell access$400(Object o, Object o2, ReferenceQueue referenceQueue) {
            return create(o, o2, referenceQueue);
        }
        
        static {
            SoftCache.ValueCell.INVALID_KEY = new Object();
            SoftCache.ValueCell.dropped = 0;
        }
    }
    
    private class Entry implements Map.Entry
    {
        private Map.Entry ent;
        private Object value;
        
        Entry(Map.Entry ent, Object value) {
            this.ent = ent;
            this.value = value;
        }
        
        @Override
        public Object getKey() {
            return this.ent.getKey();
        }
        
        @Override
        public Object getValue() {
            return this.value;
        }
        
        @Override
        public Object setValue(Object o) {
            return this.ent.setValue(SoftCache.ValueCell.access$400(this.ent.getKey(), o, SoftCache.this.queue));
        }
        
        @Override
        public boolean equals(Object o) {
            if (!(o instanceof Map.Entry)) {
                return false;
            }
            Map.Entry<Object, Object> entry = (Map.Entry)o;
            return SoftCache.access$600(this.ent.getKey(), entry.getKey()) && SoftCache.access$600(this.value, entry.getValue());
        }
        
        @Override
        public int hashCode() {
            Object key;
            return (((key = this.getKey()) == null) ? 0 : key.hashCode()) ^ ((this.value == null) ? 0 : this.value.hashCode());
        }
    }
    
    private class EntrySet extends AbstractSet
    {
        Set hashEntries;
        
        private EntrySet() {
            this.hashEntries = SoftCache.this.hash.entrySet();
        }
        
        @Override
        public Iterator iterator() {
            return new Iterator() {
                Iterator hashIterator = SoftCache.EntrySet.this.hashEntries.iterator();
                SoftCache.Entry next = null;
                
                @Override
                public boolean hasNext() {
                    while (this.hashIterator.hasNext()) {
                        Map.Entry<Object, Object> entry = (Map.Entry)this.hashIterator.next();
                        SoftCache.ValueCell valueCell = (SoftCache.ValueCell)entry.getValue();
                        Object value = null;
                        if (valueCell != null && (value = valueCell.get()) == null) {
                            continue;
                        }
                        this.next = new SoftCache.Entry(entry, value);
                        return true;
                    }
                    return false;
                }
                
                @Override
                public Object next() {
                    if (this.next == null && !this.hasNext()) {
                        throw new NoSuchElementException();
                    }
                    SoftCache.Entry next = this.next;
                    this.next = null;
                    return next;
                }
                
                @Override
                public void remove() {
                    this.hashIterator.remove();
                }
            };
        }
        
        @Override
        public boolean isEmpty() {
            return !this.iterator().hasNext();
        }
        
        @Override
        public int size() {
            int n = 0;
            Iterator<Object> iterator = this.iterator();
            while (iterator.hasNext()) {
                ++n;
                iterator.next();
            }
            return n;
        }
        
        @Override
        public boolean remove(Object o) {
            SoftCache.access$800(SoftCache.this);
            return o instanceof SoftCache.Entry && this.hashEntries.remove(((SoftCache.Entry)o).ent);
        }
    }
}
